package org.expeditee.gui;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;

import org.expeditee.core.Colour;
import org.expeditee.core.Dimension;
import org.expeditee.core.Fill;
import org.expeditee.core.GradientFill;
import org.expeditee.core.Image;
import org.expeditee.core.Point;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.GraphicsManager;
import org.expeditee.items.Item;

/**
 * Manages slide animations between frames
 * Reflection is used to call these methods, therefore any additions should only need to be added in this class.
 * No further changes to other code should be required.
 * 
 * TODO: Rework above comment. cts16
 * 
 * @author cts16
 */
public class FrameTransition
{
	/** The parameters for transition methods. */
	private static final Class<?>[] REFLECTION_PARAMS = { Image.class, AxisAlignedBoxBounds.class };
	
	/** The height of the drop-shadow below transition frames. */
	private static final int SHADOW_HEIGHT = 16;
	
	/** The colour to use for shadows. */
	private static final Colour SHADOW_COLOUR = Colour.FromRGBA255(0, 0, 0, 64);
	
	/** The default duration for transitions (in milliseconds). */
	public static final long DEFAULT_TRANSITION_DURATION = 500;
	
	/** Stores the image of the frame being transitioned away from. */
	private Image _fromFrameImage = null;
	
	/** Which transition method we are currently using. */
	private Method _method;
	
	/** How long the current transition is (in milliseconds). */
	private Long _duration = null;

	/** How far through the transition we are (in milliseconds). */
	private long _completed = 0;
	
	/** When we last updated the transition (as returned by System.currentTimeMillis()). */
	private long _lastUpdateTime = 0;
	
	/** Standard constructor. */
	public FrameTransition(Image fromFrameImage, String method)
	{
		_fromFrameImage = fromFrameImage;
		_method = reflectMethod(method);
	}
	
	/** Whether the transition has been constructed properly. */
	public boolean isValid()
	{
		return _fromFrameImage != null && _method != null;
	}
	
	/** Sets the timing data for this transition to their initial default values. */
	private void initialiseTiming()
	{
		initialiseTiming(DEFAULT_TRANSITION_DURATION);
	}
	
	/** Sets the timing data for this transition to their initial values. */
	private void initialiseTiming(long duration)
	{
		// Check if already initialised
		if (_duration != null || duration < 0) return;
		
		_duration = duration;
		_completed = 0;
		_lastUpdateTime = System.currentTimeMillis();
	}
	
	/**
	 * Updates the timing data for the transition and returns the elapsed
	 * time since the last update.
	 */
	private long updateTiming()
	{
		// No need to update if we've finished
		if (isCompleted()) return 0;
		
		long currentTime = System.currentTimeMillis();
		long deltaTime = currentTime - _lastUpdateTime;
		_completed += deltaTime;
		if (_completed > _duration) {
			long overshoot = _completed - _duration;
			_completed -= overshoot;
			currentTime -= overshoot;
			deltaTime -= overshoot;
		}
		_lastUpdateTime = currentTime;
		return deltaTime;
	}
	
	/** Gets the percentage of the transition that has elapsed. */
	private double getPercentageComplete()
	{
		if (_duration == 0.0) return 1.0;
		
		return ((double) _completed) / _duration;
	}
	
	public boolean isCompleted()
	{
		return _completed == _duration;
	}
	
	/** Gets the transition method that corresponds to the given string. */
	private Method reflectMethod(String method)
	{
		try {
			return this.getClass().getDeclaredMethod(method, REFLECTION_PARAMS);
		} catch (NoSuchMethodException | SecurityException e) {
			return null;
		}
	}
	
	/** Draws the selected transition. Returns whether the transition draw completed successfully. */
	public boolean drawTransition(Image toFrameImage, AxisAlignedBoxBounds area)
	{
		// No need to draw if not a valid transition
		if (!isValid()) return false;

		// Invoke the drawing method
		try {
			_method.invoke(this, toFrameImage, area);
			return true;
		} catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
			System.err.println("An Reflection Exception occurred trying to invoke '" + _method.getName() + "'");
			e.printStackTrace();
			return false;
		}
	}
	
	/*
	 * 
	 * Transition methods.
	 * 
	 */
	
	public void template(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();
		
		GraphicsManager g = EcosystemManager.getGraphicsManager();
	}
	
	/** Slides new frame over current from from right to left. */
	public void CoverFromRight(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();

		int frameX = area.getMinX() + (int) (area.getWidth() * (1.0 - getPercentageComplete()));
		Point topLeft = new Point(frameX, 0);
		Dimension size = area.getSize();
		
		EcosystemManager.getGraphicsManager().drawRectangle(topLeft, size, 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
		EcosystemManager.getGraphicsManager().drawImage(toImage, topLeft, size);
	}

	/** Slides new frame over current from from left to right. */
	public void CoverFromLeft(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();

		int frameX = area.getMinX() - (int) (area.getWidth() * (1.0 - getPercentageComplete()));
		Point topLeft = new Point(frameX, 0);
		Dimension size = area.getSize();
		
		EcosystemManager.getGraphicsManager().drawRectangle(topLeft, size, 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
		EcosystemManager.getGraphicsManager().drawImage(toImage, topLeft, size);
	}

	/** Another name for CoverFromRight. */
	public void Swipe(Image toImage, AxisAlignedBoxBounds area)
	{
		CoverFromRight(toImage, area);
	}

	/** Expands the new frame over the current one from the left. */
	public void WipeFromLeft(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();
		
		int cropWidth = (int) (toImage.getWidth() * getPercentageComplete());
		Dimension cropSize = new Dimension(cropWidth, toImage.getHeight());
		int drawWidth = (int) (area.getWidth() * getPercentageComplete());
		Dimension drawSize = new Dimension(drawWidth, area.getHeight());
		
		if (cropWidth == 0 || drawWidth == 0) return;
		
		EcosystemManager.getGraphicsManager().drawRectangle(area.getTopLeft(), drawSize, 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
		EcosystemManager.getGraphicsManager().drawImage(toImage, area.getTopLeft(), drawSize, 0.0, Point.ORIGIN, cropSize);
	}

	/** Expands the new frame over the current one from the right. */
	public void WipeFromRight(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();
		
		int cropWidth = (int) (toImage.getWidth() * getPercentageComplete());
		Dimension cropSize = new Dimension(cropWidth, toImage.getHeight());
		int drawWidth = (int) (area.getWidth() * getPercentageComplete());
		Dimension drawSize = new Dimension(drawWidth, area.getHeight());
		int cropX = toImage.getWidth() - cropWidth;
		Point cropTopLeft = new Point(cropX, 0);
		int drawX = area.getWidth() - drawWidth;
		Point drawTopLeft = area.getTopLeft().clone().add(drawX, 0);
		
		if (cropWidth == 0 || drawWidth == 0) return;
		
		EcosystemManager.getGraphicsManager().drawRectangle(drawTopLeft, drawSize, 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
		EcosystemManager.getGraphicsManager().drawImage(toImage, drawTopLeft, drawSize, 0.0, cropTopLeft, cropSize);
	}
		
	/**
	 * Drops new frame from top of display. Draws rectangle with gradient
	 * effect to create illusion of drop shadow.
	 */
	public void FallDown(Image toImage, AxisAlignedBoxBounds area)
	{
		initialiseTiming();
		updateTiming();
		
		int y = area.getMinY() - (int) (area.getHeight() * (1.0 - getPercentageComplete()));
		Point drawTopLeft = new Point(area.getMinX(), y);
		Point shadowTopLeft = drawTopLeft.clone().add(0, area.getHeight());
		Dimension shadowSize = new Dimension(area.getWidth(), SHADOW_HEIGHT);
		GradientFill shadowFill = new GradientFill(SHADOW_COLOUR, new Point(0, shadowTopLeft.y), Colour.TRANSPARENT, new Point(0, shadowTopLeft.y + SHADOW_HEIGHT));
		
		EcosystemManager.getGraphicsManager().drawRectangle(drawTopLeft, area.getSize(), 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
		EcosystemManager.getGraphicsManager().drawRectangle(shadowTopLeft, shadowSize, 0.0, shadowFill, null, null, null);
		EcosystemManager.getGraphicsManager().drawImage(toImage, drawTopLeft, area.getSize());
	}

	/** Fades out to the given colour and then fades in the new frame. */
	public void FadeIn(Image toImage, AxisAlignedBoxBounds area, Colour colour)
	{
		initialiseTiming();
		updateTiming();
		
		GraphicsManager g = EcosystemManager.getGraphicsManager();
		
		double progress = getPercentageComplete();
		
		// Start by fading the overlay colour in
		if (progress < 0.5) {
			float alpha = (float) (progress * 2);

			g.drawRectangle(area.getTopLeft(), area.getSize(), 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
			g.drawImage(_fromFrameImage, area.getTopLeft(), area.getSize());
			g.setCompositeAlpha(alpha);
			g.drawRectangle(area.getTopLeft(), area.getSize(), 0.0, new Fill(colour), null, null, null);
		
		// Then fade the overlay out to reveal the new frame
		} else {
			float alpha = 1.0f - (float) ((progress - 0.5) * 2);

			g.drawRectangle(area.getTopLeft(), area.getSize(), 0.0, new Fill(Item.DEFAULT_BACKGROUND), null, null, null);
			g.drawImage(toImage, area.getTopLeft(), area.getSize());
			g.setCompositeAlpha(alpha);
			g.drawRectangle(area.getTopLeft(), area.getSize(), 0.0, new Fill(colour), null, null, null);
			
		}
	}
	
	/** Fades out to black and then fades in the new frame. */
	public void FadeInBlack(Image toImage, AxisAlignedBoxBounds area)
	{
		FadeIn(toImage, area, Colour.BLACK);
	}

	/** Fades out to white and then fades in the new frame. */
	public void FadeInWhite(Image toImage, AxisAlignedBoxBounds area)
	{
		FadeIn(toImage, area, Colour.WHITE);
	}

	/** Fades the current frame out while the new frame fades in overtop. */
	public void CrossFade(Image toImage, AxisAlignedBoxBounds area)
	{ 
		// Based on: http://stackoverflow.com/questions/20346661/java-fade-in-and-out-of-images
		// TODO: make use of timer, as this example does, rather than fixed delay
		
		initialiseTiming();
		updateTiming();
		
		GraphicsManager g = EcosystemManager.getGraphicsManager();
		
		float fadeInAlpha = (float) getPercentageComplete();
		float fadeOutAlpha = 1.0f - fadeInAlpha;
		
		// Clear the screen
		g.clear(Item.DEFAULT_BACKGROUND);
		
		// Draw the current frame with decreasing alpha
		g.setCompositeAlpha(fadeOutAlpha);
		g.drawImage(_fromFrameImage, area.getTopLeft(), area.getSize());
		
		// Draw the in-image with increasing alpha
		g.setCompositeAlpha(fadeInAlpha);
		g.drawImage(toImage, area.getTopLeft(), area.getSize());
	}
}
